/* Copyright is licensed under GNU LGPL.                 by IJ. Wang 2006

  Build: make chk_array
*/
#include "wy_array.h"
#include "wy_uty.h"
#include "wystr.h"
#include "wytimespec.h"
#include "wydate.h"
#include "wyfilestat.h"
#include "wydirent.h"
#include "wytermios.h"
#include "wyselect.h"
#include "wy_thread.h"
#include "ck_array.h"

// Test object as element of Wy_Array
class TestBase {
    const int& _ref;
  public:
    TestBase(const int& v) : _ref(v) {};
    TestBase(TestBase& src) : _ref(src._ref) {};
    TestBase(TestBase& src, Wy::ByMove_t) : _ref(src._ref) {};
    const int& ref(void) const { return _ref; };
};

class Element : TestBase {
    int _v;
    const int _cpycnt;         // number that *this is copy constructed
    static int _objcnt;

  public:
   Element() : TestBase(_v),_v(0),_cpycnt(1) {
     ++_objcnt; 
   };
   Element(const Element& s) : TestBase(_v), _v(s._v), _cpycnt(s._cpycnt+1) {
     ++_objcnt; 
   };

   // [P] when base is move constructed, s is move destructed from the base,
   //     which is the reverse order of normal object destruction
   //
   Element(Element& s, Wy::ByMove_t)
       : TestBase(s,Wy::ByMove), _v(s._v), _cpycnt(s._cpycnt)
      { /* No ctor/dtor semantics:  ++_objcnt; s.~Element(); */ };

   Element(int v) : TestBase(_v),_v(v),_cpycnt(1) {
      ++_objcnt; 
   };
   ~Element() { 
      --_objcnt; 
   };
   bool is_default(void) const { return _v==0; };
   int value(void) const { return _v; };
   int copy_count(void) const { return _cpycnt; };
   void reset(int v) { _v=v; };
   const Element& operator =(int v) { _v=v; return *this; };
   const Element& operator =(const Element& rhs)
         { _v=rhs._v; return *this; };
   bool operator <(const Element& rhs) const { return this->_v<rhs._v; };
   bool operator ==(const Element& rhs) const { return this->_v==rhs._v; };
   bool operator !=(const Element& rhs) const { return this->_v!=rhs._v; };
   
   void swap(Element& v)
       {
        int tmp(_v);
        _v=v._v;
        v._v=tmp;
       };

   static int obj_cnt(void) { return _objcnt; };
};
int Element::_objcnt=0;

// Copy constructor throw Reply for test
//
class ErrCpyCtor {
    int _v;
  public:
    WY_THROW_REPLY;
    ErrCpyCtor() : _v(1000) {};
    ErrCpyCtor(const ErrCpyCtor& rhs) : _v(rhs._v) { throw Reply(); };
    ErrCpyCtor(ErrCpyCtor& rhs,Wy::ByMove_t) : _v(rhs._v) {};
    bool operator== (const ErrCpyCtor& rhs) const
         { return _v==rhs._v; };
    bool operator!= (const ErrCpyCtor& rhs) const
         { return _v!=rhs._v; };
};

#define CHK_DEFAULT WY__CK_ARRAY_CHK_DEFAULT

static void t0(void)
{
};

static void t1()
{
 if(Element::obj_cnt()!=0) {
   WY_THROW( WyRet() );
 }
 if((Wy_Array<Element>::min_capacity()<1) ||
    (Wy_Array<Element>::min_capacity()>=Wy_Array<Element>::max_capacity())) {
   WY_THROW( WyRet() );
 }
 if(Wy_Array<Element>::max_capacity()>INT_MAX/sizeof(Element)-1) {
   WY_THROW( WyRet() );
 }
 WyRet r;
 Wy_Array<Element> tmp;

 if(Element::obj_cnt()!=0) {
   WY_THROW( WyRet() );
 }
 if(tmp._capacity()!=4) {  // predefined value
   WY_THROW( WyRet() );
 }
 if(tmp.is_default()==false) {
   WY_THROW( WyRet() );
 }
 if(tmp!=tmp) {
   WY_THROW( WyRet() );
 }
 if(tmp==tmp) {
 } else {
   WY_THROW( WyRet() );
 }
 CHK_DEFAULT(tmp);

 // chk ctor(size_t,const Element&)
 {
  Wy_Array<Element> tmp(1,Element());
  if(tmp._capacity()<tmp.min_capacity()) {
    WY_THROW( WyRet() );
  };
  if(tmp.size()!=1) {
    WY_THROW( WyRet() );
  };
  for(size_t i=0; i<tmp.size(); ++i) {
    if(tmp[i].is_default()==false) {
      WY_THROW( WyRet() );
    };
  }

  Wy_Array<Element> tmp2(tmp.min_capacity()+1,Element());
  if(tmp2._capacity()<tmp2.min_capacity()+1) {
    WY_THROW( WyRet() );
  };
  if(tmp2.size()<tmp2.min_capacity()+1) {
    WY_THROW( WyRet() );
  };
  for(size_t i=0; i<tmp2.size(); ++i) {
    if(tmp2[i].is_default()==false) {
      WY_THROW( WyRet() );
    };
  }

  size_t ncap( 2*Wy_Array<Element>::min_capacity() );
  Wy_Array<Element> tmp3(ncap,Element(100));
  if(tmp3._capacity()<ncap) {
    WY_THROW( WyRet() );
  };
  for(size_t i=0; i<tmp3.size(); ++i) {
    if(tmp3[i].is_default()==true) {
      WY_THROW( WyRet() );
    };
    if(tmp3[i]!=100) {
      WY_THROW( WyRet() );
    };
  }

 }

 // chk copy ctor (1)
 {
  Wy_Array<Element> tmp2(tmp);
  CHK_DEFAULT(tmp2);
 }

 // chk error push_back()
 {
  Wy_Array<ErrCpyCtor> tmp1;
  Wy_Array<ErrCpyCtor> tmp2(tmp1);
  if(tmp2!=tmp1) {
    WY_THROW( WyRet() );
  }
  try {
    tmp2.push_back( ErrCpyCtor() );
    WY_THROW( WyRet() );
  }
  catch(const ErrCpyCtor::Reply& e) {
    if(tmp2!=tmp1) {
      WY_THROW( WyRet() );
    }
    if(e.is_default()==false) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };
 }

 // chk push_back
 {
   tmp.push_back( 53 );
   if(tmp.is_default()==true) {
     WY_THROW( WyRet() );
   }
   if(tmp.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].value()!=53) {
     WY_THROW( WyRet() );
   }
   if(Element::obj_cnt()!=1) {
     Wy::cout << "objcnt= " << Element::obj_cnt() << '\n';
     WY_THROW( WyRet() );
   }

   tmp.push_back( 22 );
   if(tmp.is_default()==true) {
     WY_THROW( WyRet() );
   }
   if(tmp.size()!=2) {
     WY_THROW( WyRet() );
   }
   if(tmp[1].value()!=22) {
     WY_THROW( WyRet() );
   }

   tmp.push_back(tmp.back());
   if(tmp.is_default()==true) {
     WY_THROW( WyRet() );
   }
   if(tmp.size()!=3) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].value()!=53) {
     WY_THROW( WyRet() );
   }
   if(tmp[1].value()!=22) {
     WY_THROW( WyRet() );
   }
   if(tmp[2].value()!=22) {
     WY_THROW( WyRet() );
   }

   tmp.push_back(tmp[0]);
   if(tmp.back()!=tmp[0]) {
     WY_THROW( WyRet() );
   }
   tmp.pop_back();
   tmp.pop_back();

   //if(Element::obj_cnt()!=0) {
   //  Wy::cout << "objcnt= " << Element::obj_cnt() << '\n';
   //  WY_THROW( WyRet() );
   //}
   const size_t org_size=tmp.size();
   const size_t new_size=tmp._capacity()+1;
   const int o_cpycnt= tmp[0].copy_count();
   while(tmp.size()<new_size) {
     tmp.push_back(44);
   }
   if(o_cpycnt!=tmp[0].copy_count()) {
     WY_THROW( WyRet() );   // tmp[0] was modified by the copy ctor
   }
   tmp.resize(org_size);
 }

 // chk copy ctor (2)
 {
  Wy_Array<Element> tmp2(tmp);
  if(tmp2.size()!=tmp.size()) {
    WY_THROW( WyRet() );
  }
  for(size_t i=0; i<tmp.size(); ++i) {
    if(tmp[i]!=tmp2[i]) {
      WY_THROW( WyRet() );
    }
  }
 }

 if(Element::obj_cnt()!=2) {
   Wy::cout << "objcnt= " << Element::obj_cnt() << '\n';
   WY_THROW( WyRet() );
 }

 // chk push_back/copy constructor (reallocate)
 tmp.reset(); 
 CHK_DEFAULT(tmp);
 for(int i=0; i<1000000; ++i) {
   tmp.push_back(i);
   if(tmp._capacity()<tmp.size()) {
     WY_THROW( WyRet() );
   }
   if((int)tmp.size()!=i+1) {
     WY_THROW( WyRet() );
   }
   if(tmp[i].value()!=i) {
     WY_THROW( WyRet() );
   }
 }
 {
  const Element* p=tmp.end();
  if(p!=tmp.begin()) {
    do {
      --p;
    } while(p!=tmp.begin());
  }
 }
 {
   Wy_Array<Element> t2(tmp);

   if(t2.size()!=tmp.size()) {
     WY_THROW( WyRet() );
   }
   for(size_t i=0; i<t2.size(); ++i) {
     if(tmp[i].value()!=t2[i].value()) {
       WY_THROW( WyRet() );
     }
   }
 }
 while(tmp.size()) {
   tmp.pop_back();
 };
 CHK_DEFAULT(tmp);

 // chk front/back/
 {
  tmp.reset();
  CHK_DEFAULT(tmp);
  tmp.push_back(44);
  if(tmp.front()!=44) {
    WY_THROW( WyRet() );
  }
  if(tmp.back()!=44) {
    WY_THROW( WyRet() );
  }

  tmp.push_back(55);
  if(tmp.front()!=44) {
    WY_THROW( WyRet() );
  }
  if(tmp.back()!=55) {
    WY_THROW( WyRet() );
  }

  tmp.front()=1;
  tmp.back()=2;
  if(tmp.front()!=1) {
    WY_THROW( WyRet() );
  }
  if(tmp.back()!=2) {
    WY_THROW( WyRet() );
  }
 }

 // chk _reserve
 {
   Wy_Array<Element> tmp;
   const size_t NCap=1000;

   CHK_DEFAULT(tmp);
   if(tmp._capacity()==NCap) {
     WY_THROW( WyRet() );      // may be ok (default capacity not specified)
   }
   if((r=tmp._reserve(NCap))!=OK) {
     WY_THROW(r);
   }
   if(tmp._capacity()!=NCap) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tmp);

   if((r=tmp._reserve(NCap-1))!=OK) {
     WY_THROW(r);
   }
   if(tmp._capacity()!=NCap) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tmp);
 }

 // chk erase(size_t,size_t)
 {
   Wy_Array<Element> tmp;

   CHK_DEFAULT(tmp);
   tmp.push_back(10);
   tmp.push_back(11);
   tmp.push_back(12);
   tmp.push_back(13);
   if(tmp.size()!=4) {
     WY_THROW( WyRet() );
   }

   Wy_Array<Element> tmp_cpy(tmp);
   if(tmp!=tmp_cpy) {
     WY_THROW( WyRet() );
   }
   try {
     tmp.erase(tmp.size()+1,1);
     WY_THROW( WyRet() );
   }
   catch(const Wy_Array<Element>::Reply& e) {
     if(tmp!=tmp_cpy) {
       WY_THROW( WyRet() );
     }
     if(e!=Wym_EINVAL) {
       WY_THROW( WyRet(e) );
     }
   }
   catch(...) {
     WY_THROW( WyRet() );
   };

   tmp.erase(tmp.size(),1);
   tmp.erase(3,0);
   if(tmp.size()!=4) {
     WY_THROW( WyRet() );
   }
   tmp.erase(3,1);

   tmp_cpy.reset(tmp);
   if(tmp!=tmp_cpy) {
     WY_THROW( WyRet() );
   }
   try {
     tmp.erase(4,1);
     WY_THROW( WyRet() );
   }
   catch(const Wy_Array<Element>::Reply& e) {
     if(tmp!=tmp_cpy) {
       WY_THROW( WyRet() );
     }
     if(e!=Wym_EINVAL) {
       WY_THROW( WyRet(e) );
     }
   }
   catch(...) {
     WY_THROW( WyRet() );
   };

   tmp.erase(1,2);
   if(tmp.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp[0]!=10) {
     WY_THROW( WyRet() );
   }
 }

 // chk swap
 {
   Wy_Array<Element> tmp1,tmp2;
   CHK_DEFAULT(tmp1);
   CHK_DEFAULT(tmp2);

   tmp1.push_back(1);
   tmp1.swap(tmp2);
   CHK_DEFAULT(tmp1);
   if(tmp2[0].value()!=1) {
    WY_THROW( WyRet() );
   }

   tmp2.push_back(2);
   tmp2[0].swap(tmp2[1]);
   if(tmp2[0].value()!=2) {
    WY_THROW( WyRet() );
   }
   if(tmp2[1].value()!=1) {
    WY_THROW( WyRet() );
   }

   tmp2.reset(tmp1);
   tmp2.swap(tmp2);    // self-swap
   if(tmp2!=tmp1) {
    WY_THROW( WyRet() );
   }
 }

 // chk reset(const Wy_Array&)
 {
   Wy_Array<Element> tmp1;
   CHK_DEFAULT(tmp1);
   Wy_Array<Element> tmp2;
   CHK_DEFAULT(tmp2);

   tmp2.reset(tmp2);
   CHK_DEFAULT(tmp2);
   tmp2.reset(tmp1);
   CHK_DEFAULT(tmp2);

   tmp1.push_back(8);
   tmp2.reset(tmp1);
   if(tmp2!=tmp1) {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }

   tmp1.push_back(7);
   tmp2.reset(tmp1);
   if(tmp2==tmp1) {
   } else {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }

   tmp2.reset(tmp2);   // self-reset
   if(tmp2!=tmp1) {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }
 }

 // chk operator= (const Wy_Array&)
 {
   Wy_Array<Element> tmp1;
   CHK_DEFAULT(tmp1);
   Wy_Array<Element> tmp2;
   CHK_DEFAULT(tmp2);

   if(&(tmp2=tmp1)!=&tmp2) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tmp2);
   tmp2=tmp2;
   CHK_DEFAULT(tmp2);

   tmp1.push_back(8);
   if(&(tmp2=tmp1)!=&tmp2) {
     WY_THROW( WyRet() );
   }
   if(tmp2!=tmp1) {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }

   tmp1.push_back(7);
   tmp2=tmp1;
   if(tmp2==tmp1) {
   } else {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }

   tmp2=tmp2;   // self-reset
   if(tmp2!=tmp1) {
     WY_THROW( WyRet() );
   }
   if(tmp2._capacity()<tmp2.min_capacity()) {
     WY_THROW( WyRet() );
   }
 }

 // chk insert
 {
   Wy_Array<Element> tmp;
   CHK_DEFAULT(tmp);

   tmp.insert(0,4);
   if(tmp.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].value()!=4) {
     WY_THROW( WyRet() );
   }

   tmp.insert(0,3);
   if(tmp.size()!=2) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].value()!=3) {
     WY_THROW( WyRet() );
   }
   if(tmp[1].value()!=4) {
     WY_THROW( WyRet() );
   }

   tmp.insert(1,5);
   if(tmp.size()!=3) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].value()!=3) {
     WY_THROW( WyRet() );
   }
   if(tmp[1].value()!=5) {
     WY_THROW( WyRet() );
   }
   if(tmp[2].value()!=4) {
     WY_THROW( WyRet() );
   }

   Wy_Array<Element> tmp_cpy(tmp);
   if(tmp!=tmp_cpy) {
     WY_THROW( WyRet() );
   }
   try {
     tmp.insert(99,5);
   }
   catch(const Wy_Array<Element>::Reply& e) {
     if(tmp!=tmp_cpy) {
       WY_THROW( WyRet() );
     }
     if(e!=Wym_EINVAL) {
       WY_THROW( WyRet(e) );
     }
   }
   catch(...) {
     WY_THROW( WyRet() );
   };

   tmp.reset();
   CHK_DEFAULT(tmp);
   for(int i=0; i<100; ++i) {
     tmp.insert(i,i);
     if(tmp[i].value()!=i) {
       WY_THROW( WyRet() );
     }
     if(tmp.size()!=size_t(i+1)) {
       WY_THROW( WyRet() );
     }
   }
   for(int i=0; i<100; ++i) {
     if(tmp[i].value()!=i) {
       WY_THROW( WyRet() );
     }
   }

   tmp.reset();
   CHK_DEFAULT(tmp);
   tmp.insert(0,0);
   tmp.insert(1,1);
   tmp.insert(2,2);           // 012
   tmp.insert(1,tmp.back());  // 0212
   tmp.insert(3,tmp.back());  // 02122
   if(tmp[0]!=0) {
     WY_THROW( WyRet() );
   }
   if(tmp[1]!=2) {
     WY_THROW( WyRet() );
   }
   if(tmp[2]!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp[3]!=2) {
     WY_THROW( WyRet() );
   }
   if(tmp[4]!=2) {
     WY_THROW( WyRet() );
   }
 }

 // chk resize(size_t)
 {
   Wy_Array<Element> tmp;
   CHK_DEFAULT(tmp);

   tmp.resize(0);

   CHK_DEFAULT(tmp);
   tmp.resize(1);
   if(tmp.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp[0].is_default()==false) {
     WY_THROW( WyRet() );
   }
   tmp.push_back(1);
   tmp.resize(3);
   if(tmp.size()!=3) {
     WY_THROW( WyRet() );
   }

   Wy_Array<Element> tmp2(tmp);
   tmp.resize(99);
   if(tmp.size()!=99) {
     WY_THROW( WyRet() );
   }
   if(tmp[1]!=1) {
     WY_THROW( WyRet() );
   }
   tmp.resize(tmp2.size());
   if(tmp!=tmp2) {
     WY_THROW( WyRet() );
   }

   Wy_Array<Element> tmp_cpy(tmp);
   if(tmp!=tmp_cpy) {
     WY_THROW( WyRet() );
   }
   try {
     tmp.resize(tmp.max_capacity()+1);
     WY_THROW( WyRet() );
   }
   catch(const Wy_Array<Element>::Reply& e) {
     if(tmp!=tmp_cpy) {
       WY_THROW( WyRet() );
     }
     if(e!=Wym_EFBIG) {
       WY_THROW( WyRet(e) );
     }
   }
   catch(...) {
     WY_THROW( WyRet() );
   };
 }

 // chk replace(size_t,size_t,size_t,const T&)
 {
   const int ObjCnt=Element::obj_cnt();
   Wy_Array<Element> tmp;
   CHK_DEFAULT(tmp);
   try {
     tmp.replace(1,1,1,Element());
     WY_THROW( WyRet() );
   }
   catch(const Wy_Array<Element>::Reply& e) {
     if(e!=Wym_EINVAL) {
       WY_THROW( WyRet(e) );
     }
   };
   CHK_DEFAULT(tmp);
   if(ObjCnt!=Element::obj_cnt()) {
     WY_THROW( WyRet() );
   }

   // loop test, so the second loop wont reallocate buffer
   //
   for(int loopcnt=0; loopcnt<3; ++loopcnt) {
     tmp.replace(0,1,0,Element());
     CHK_DEFAULT(tmp);
     if(ObjCnt!=Element::obj_cnt()) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(0,1,1,Element(5));
     if(tmp.size()!=1) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=5) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+1) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(0,0,1,Element(4));
     if(tmp.size()!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+2) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(2,0,1,Element(6));
     if(tmp.size()!=3) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+3) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=5) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=6) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(1,1,1,Element(2));
     if(tmp.size()!=3) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+3) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=6) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(1,1,2,Element(2));
     if(tmp.size()!=4) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+4) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[3]!=6) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(0,1,2,Element(4));
     if(tmp.size()!=5) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+5) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[3]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[4]!=6) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(4,1,2,Element(6));
     if(tmp.size()!=6) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+6) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[3]!=2) {
       WY_THROW( WyRet() );
     }
     if(tmp[4]!=6) {
       WY_THROW( WyRet() );
     }
     if(tmp[5]!=6) {
       WY_THROW( WyRet() );
     }
  
     tmp.replace(1,4,1,Element(5));
     if(tmp.size()!=3) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+3) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=4) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=5) {
       WY_THROW( WyRet() );
     }
     if(tmp[2]!=6) {
       WY_THROW( WyRet() );
     }

     tmp.replace(0,1,0,Element(5));
     if(tmp.size()!=2) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+2) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=5) {
       WY_THROW( WyRet() );
     }
     if(tmp[1]!=6) {
       WY_THROW( WyRet() );
     }

     tmp.replace(1,4,0,Element(5));
     if(tmp.size()!=1) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+1) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=5) {
       WY_THROW( WyRet() );
     }

     tmp.replace(1,4,0,Element(5));
     if(tmp.size()!=1) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+1) {
       WY_THROW( WyRet() );
     }
     if(tmp[0]!=5) {
       WY_THROW( WyRet() );
     }

     tmp.replace(0,4,0,Element(5));
     if(tmp.size()!=0) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+0) {
       WY_THROW( WyRet() );
     }
     tmp.replace(0,4,0,Element(5));
     if(tmp.size()!=0) {
       WY_THROW( WyRet() );
     }
     if(Element::obj_cnt()!=ObjCnt+0) {
       WY_THROW( WyRet() );
     }

     tmp.reset();
   }

   tmp.reset();
   tmp.push_back(11);
   for(int loopcnt=0; loopcnt<100; ++loopcnt) {
     tmp.push_back( tmp.back() );
     if(tmp.back()!=11) {
       WY_THROW( WyRet() );
     }
   }
   for(size_t i=0; i<tmp.size(); ++i) {
     if(tmp[i]!=11) {
       WY_THROW( WyRet() );
     }
   }
 }
};

static void t2()
{
 WyRet r;
 {
  Wy_Array<Element> tmp1;
  CHK_DEFAULT(tmp1);
  
  Wy_Array< Wy_Array<Element> > tmp_a;
  CHK_DEFAULT(tmp_a);

  tmp_a.push_back(tmp1);
  if(tmp_a.size()!=1) {
    WY_THROW( WyRet() );
  }

  for(int i=0; i<1000; ++i) {
    tmp_a.push_back(tmp1);
    if(tmp_a[0]!=tmp1) {
      WY_THROW( WyRet() );
    }
    if(tmp_a.back()!=tmp1) {
      WY_THROW( WyRet() );
    }
  }
  for(; tmp_a.size(); ) {
    if(tmp_a.back()!=tmp1) {
      WY_THROW( WyRet() );
    }
    tmp_a.pop_back();
  }
  CHK_DEFAULT(tmp_a);
 }
};

template<typename T>
static void t_qsort(Wy_Array<T>& v, size_t start, size_t end)
{
 const size_t VLen( end-start );
 if(VLen<=1) {
   return;
 }
 if(VLen==2) {
   if(v[start+1]<v[start]) {
     v[start].swap(v[start+1]);
   }
   return;
 }
 size_t ge_bgn=start;
 const T pe(v[ge_bgn]);
 for(size_t s=start+1; s<end; ++s) {
   if(v[s]<pe) {
     v[ge_bgn].swap(v[s]);
     ++ge_bgn;
   }
 }
 if(ge_bgn==start) {
   t_qsort(v,start+1,end);
   return;
 }
 t_qsort(v,start,ge_bgn);
 t_qsort(v,ge_bgn,end);
};

static void t_sort(void)
{
 WyRet r;
 Wy_Array<Element> v;
 v.push_back(2);
 v.push_back(2);
 v.push_back(1);
 v.push_back(2);
 v.push_back(3);
 v.push_back(2);
 v.push_back(5);
 v.push_back(4);
 v.push_back(2);
 v.push_back(4);
 v.push_back(2);

 t_qsort(v,0,v.size());
 for(size_t i=1; i<v.size(); ++i) {
   if(v[i]<v[i-1]) {
     WY_THROW( WyRet() );
   }
 }
};

// EleType is arithmetic type
template<typename EleType>
static void ck_array_value(void)
{
 const EleType DefaultElem( EleType(0) );

 {
  Wy_Array<EleType> tmp;
  if(tmp.begin()!=tmp.end()) {
    WY_THROW( WyRet() );
  }
  if(tmp.size()!=0) {
    WY_THROW( WyRet() );
  }
  if(tmp._capacity()!=Wy_Array<EleType>::min_capacity()) {
    WY_THROW( WyRet() );
  }
  CHK_DEFAULT(tmp);
 }

 // chk Wy_Array(const Wy_Array&)
 {
  Wy_Array<EleType> tmp1;
  CHK_DEFAULT(tmp1);
  Wy_Array<EleType> tmp2(tmp1);
  CHK_DEFAULT(tmp2);

  tmp2.push_back(EleType());
  Wy_Array<EleType> tmp3(tmp2);
  if(tmp2!=tmp3) {
    WY_THROW( WyRet() );
  }
  if(tmp3.size()!=1) {
    WY_THROW( WyRet() );
  }

  for(int i=0; i<100; ++i) {
    tmp2.push_back(EleType());
  }
  Wy_Array<EleType> tmp4(tmp2);
  if(tmp4!=tmp2) {
    WY_THROW( WyRet() );
  }
  if(tmp4.size()!=tmp2.size()) {
    WY_THROW( WyRet() );
  }
  if(tmp4._capacity()<tmp4.size()) {
    WY_THROW( WyRet() );
  }
 }

 // chk Wy_Array(size_t,const EleType&)
 {
  EleType ele;
  Wy_Array<EleType> tmp1(size_t(40),ele),tmp2(tmp1);
  if(tmp1.size()!=40) {
    WY_THROW( WyRet() );
  }
  if(tmp2.size()!=tmp1.size()) {
    WY_THROW( WyRet() );
  }

  Wy_Array<EleType*> tmp3(size_t(40),&ele),tmp4(tmp3);
  if(tmp3.size()!=40) {
    WY_THROW( WyRet() );
  }
  if(tmp4.size()!=tmp3.size()) {
    WY_THROW( WyRet() );
  }

  Wy_Array<const EleType*> tmp5(size_t(40),&ele),tmp6(tmp5);
  if(tmp5.size()!=40) {
    WY_THROW( WyRet() );
  }
  if(tmp6.size()!=tmp5.size()) {
    WY_THROW( WyRet() );
  }
 
 }

 // chk err front()
 {
  Wy_Array<EleType> tmp1;
  CHK_DEFAULT(tmp1);

  EleType tmp_e(DefaultElem);
  if(tmp_e!=DefaultElem) {
    WY_THROW( WyRet() );
  }
  try {
    tmp_e=tmp1.front();
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(tmp_e!=DefaultElem) {
      WY_THROW( WyRet() );
    }
    if(e!=Wym_ENOENT) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };

  CHK_DEFAULT(tmp1);
  try {
    tmp1.front()=DefaultElem;  // EleType has reset() defined
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    CHK_DEFAULT(tmp1);
    if(e!=Wym_ENOENT) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };
 }

 // chk err back()
 {
  Wy_Array<EleType> tmp1;
  CHK_DEFAULT(tmp1);
  try {
    EleType tmp_e( tmp1.back() );
    tmp_e=EleType(0);    // supress warning
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(e!=Wym_ENOENT) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };

  try {
    tmp1.back()=DefaultElem;  // EleType has operator= defined
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    CHK_DEFAULT(tmp1);
    if(e!=Wym_ENOENT) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };
 }

 // chk err operator[]
 {
  Wy_Array<EleType> tmp1;
  CHK_DEFAULT(tmp1);
  try {
    EleType tmp_e( tmp1[0] );
    tmp_e=DefaultElem;       // supress warning message
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    CHK_DEFAULT(tmp1);
    if(e!=Wym_EINVAL) {
      WY_THROW( WyRet(e) );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };

  tmp1.push_back( EleType() );
  tmp1[0]=DefaultElem;
  if(tmp1.size()!=1) {
    WY_THROW( WyRet() );
  }
  if(tmp1[0]!=DefaultElem) {
    WY_THROW( WyRet() );
  }
  try {
    tmp1[tmp1.size()]=DefaultElem;
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(e!=Wym_EINVAL) {
      WY_THROW( WyRet(e) );
    }
    if(tmp1.size()!=1) {
      WY_THROW( WyRet() );
    }
    if(tmp1[0]!=DefaultElem) {
      WY_THROW( WyRet() );
    }
  }
  catch(...) {
    WY_THROW( WyRet() );
  };
 }

 // chk push_back
 {
  Wy_Array<EleType> tmp;
  CHK_DEFAULT(tmp);
  size_t cap( (tmp._capacity()<<1)+1 );
  for(size_t i=0; i<cap; ++i) {
    tmp.push_back( EleType() );
  }
  if(tmp.size()!=cap) {
    WY_THROW( WyRet() );
  }
  for(size_t i=0; i<tmp.size(); ++i) {
    if(tmp[i]!=DefaultElem) {
      WY_THROW( WyRet() );
    }
  }
  tmp.reset();
  CHK_DEFAULT(tmp);
  tmp.push_back( EleType() );
  tmp.push_back( EleType(1) );
  tmp.push_back( tmp.back() );
  if(tmp.size()!=3) {
    WY_THROW( WyRet() );
  }
  if(tmp[0]!=DefaultElem) {
    WY_THROW( WyRet() );
  }
  if(tmp[1]!=EleType(1)) {
    WY_THROW( WyRet() );
  }
  if(tmp[2]!=EleType(1)) {
    WY_THROW( WyRet() );
  }
 }

 // chk pop_back
 {
  Wy_Array<EleType> tmp;
  CHK_DEFAULT(tmp);
  tmp.pop_back();
  CHK_DEFAULT(tmp);

  size_t cap( (tmp._capacity()<<1)+1 );
  for(size_t i=0; i<cap; ++i) {
    tmp.push_back( EleType() );
  }
  if(tmp.size()!=cap) {
    WY_THROW( WyRet() );
  }
  while(tmp.size()) {
    tmp.pop_back();
  };
  if(tmp.is_default()==false) {
    WY_THROW( WyRet() );
  }
 }

 // chk erase
 {
  Wy_Array<EleType> tmp;
  CHK_DEFAULT(tmp);
  tmp.erase(0,1);
  CHK_DEFAULT(tmp);

  size_t cap( (tmp._capacity()<<1)+1 );
  for(size_t i=0; i<cap; ++i) {
    tmp.push_back( EleType() );
  }
  if(tmp.size()!=cap) {
    WY_THROW( WyRet() );
  }
  size_t rsize=tmp.size();
  tmp.erase(tmp.size(),100);
  if(tmp.size()!=rsize) {
    WY_THROW( WyRet() );
  }
  tmp.erase(tmp.size()-1,0);
  if(tmp.size()!=rsize) {
    WY_THROW( WyRet() );
  }
  tmp.erase(0,tmp.size());
  if(tmp.size()!=0) {
    WY_THROW( WyRet() );
  }
  rsize=tmp.size();

  try {
    tmp.erase(tmp.size()+1,1);
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(e!=Wym_EINVAL) {
      WY_THROW( WyRet() );
    }
  };
  if(tmp.size()!=rsize) {
    WY_THROW( WyRet() );
  }
 }

 // chk insert
 {
  Wy_Array<EleType> tmp;
  CHK_DEFAULT(tmp);
  tmp.insert(0,DefaultElem);
  if(tmp.size()!=1) {
    WY_THROW( WyRet() );
  }
  tmp.insert(0,DefaultElem);
  if(tmp.size()!=2) {
    WY_THROW( WyRet() );
  }
  tmp.insert(0,DefaultElem);
  if(tmp.size()!=3) {
    WY_THROW( WyRet() );
  }
  tmp.insert(1,DefaultElem);
  if(tmp.size()!=4) {
    WY_THROW( WyRet() );
  }
  tmp.insert(tmp.size()-1,EleType(5));  // xxx5x
  if(tmp.size()!=5) {
    WY_THROW( WyRet() );
  }
  tmp.insert(tmp.size(),EleType(6));    // xxx5x6
  if(tmp.size()!=6) {
    WY_THROW( WyRet() );
  }
  tmp.insert(tmp.size()-2,tmp.back());  // xxx56x6
  if(tmp.size()!=7) {
    WY_THROW( WyRet() );
  }
  if(tmp[3]!=EleType(5)) {
    WY_THROW( WyRet() );
  }
  if(tmp[4]!=EleType(6)) {
    WY_THROW( WyRet() );
  }
  if(tmp[5]!=DefaultElem) {
    WY_THROW( WyRet() );
  }
  if(tmp[6]!=EleType(6)) {
    WY_THROW( WyRet() );
  }

  try {
    tmp.insert(tmp.size()+1,EleType());
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(e!=Wym_EINVAL) {
      WY_THROW( WyRet() );
    }
  };
  if(tmp.size()!=7) {
    WY_THROW( WyRet() );
  }
 }

 // chk resize
 {
  Wy_Array<EleType> tmp;
  CHK_DEFAULT(tmp);
  tmp.resize(0);
  CHK_DEFAULT(tmp);

  tmp.resize(1);
  if(tmp.size()!=1) {
    WY_THROW( WyRet() );
  }
  tmp.resize(2);
  if(tmp.size()!=2) {
    WY_THROW( WyRet() );
  }
  tmp.resize(100);
  if(tmp.size()!=100) {
    WY_THROW( WyRet() );
  }
  tmp.resize(101);
  if(tmp.size()!=101) {
    WY_THROW( WyRet() );
  }
  tmp.resize(90);
  if(tmp.size()!=90) {
    WY_THROW( WyRet() );
  }
  try {
    tmp.resize(tmp.max_capacity()+1);
    WY_THROW( WyRet() );
  }
  catch(const typename Wy_Array<EleType>::Reply& e) {
    if(e!=Wym_EFBIG) {
      WY_THROW( WyRet() );
    }
  };
  if(tmp.size()!=90) {
    WY_THROW( WyRet() );
  }
  tmp.resize(1);
  if(tmp.size()!=1) {
    WY_THROW( WyRet() );
  }
  tmp.resize(0);
  if(tmp.size()!=0) {
    WY_THROW( WyRet() );
  }
  CHK_DEFAULT(tmp);
 }

};

// chk Wy_Array<WyCSeg>, for WyCSeg not defined operator==
static void ck_array_cseg(void)
{
 Wy_Array<WyCSeg> tmp;
 if(tmp.is_default()==false) {
   WY_THROW( WyRet() );
 }

 tmp.push_back(WyCSeg());
 if(tmp.size()!=1) {
   WY_THROW( WyRet() );
 }

};

enum ENUM { A,B,C };
template<>
struct Wy_Array<ENUM> : Wy__ArrayValue<ENUM> {
  Wy_Array() : Wy__ArrayValue<ENUM>() {};
  Wy_Array(size_t s, const ENUM& elem) : Wy__ArrayValue<ENUM>(s,elem) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<ENUM>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<ENUM>(p,t) {};
};

static void ck_2d(void)
{
 Wy_Array< Wy_Array<int> > a(10, Wy_Array<int>(10,int()));
 if(a.size()!=10) {
   WY_THROW( WyRet() );
 }
 for(size_t i=0; i<a.size(); ++i) {
   if(a[i].size()!=10) {
     WY_THROW( WyRet() );
   }
   for(size_t j=0; j<a[i].size(); ++j) {
     //a[i][j]=i*j;
   }
 }
};

int main(void)
try {
 Wy::cout << "Checking wy_array.h ...\n";
 t0();

 int cobjs( Element::obj_cnt() );
 t1();
 if(Element::obj_cnt()!=cobjs) {
   Wy::cout << "objcnt before t1()= " << cobjs
            << ", after t1()= " << Element::obj_cnt() << '\n';
   WY_THROW( WyRet() );
 }

 cobjs= Element::obj_cnt();
 t2();
 if(Element::obj_cnt()!=cobjs) {
   Wy::cout << "objcnt before t2()= " << cobjs
            << ", after t2()= " << Element::obj_cnt() << '\n';
   WY_THROW( WyRet() );
 }

 t_sort();
 ck_array< Wy_Array<Element> >();
 ck_array< Wy_Array< Wy_Array<Element> > >();
 ck_array<WyReply>();
 ck_array<WyRet>();
 ck_array<WyStr>();
 ck_array<WyTimeSpec>();
 ck_array<WyDate>();
 ck_array<WyFileStat>();
 ck_array<WyFileHandle>();
 ck_array<WyDirEnt>();
 ck_array<WyTermios>();
 //ck_array<Wy_FdSet>();
 ck_array<Wy_ThreadID>();

 ck_array_cseg();
 ck_array_value<char>(); 
 ck_array_value<bool>();
 ck_array_value<signed char>();
 ck_array_value<signed short>();
 ck_array_value<signed int>();
 ck_array_value<signed long int>();
 ck_array_value<signed long long int>();
 ck_array_value<unsigned char>();
 ck_array_value<unsigned short>();
 ck_array_value<unsigned int>();
 ck_array_value<unsigned long int>();
 ck_array_value<unsigned long long int>();
 ck_array_value<float>();
 ck_array_value<double>();
 ck_array_value<long double>();

 ck_array_value<char*>();
 ck_array_value<const int*>();
 ck_array_value<float*>();
 ck_array_value<Element*>();
 ck_array_value<const Element*>();
 ck_array_value<int(*)(int)>();

 ck_array_value<ENUM>();

 ck_array<bool>();
 ck_array<signed int>();
 ck_array<unsigned int>();
 ck_array<float>();
 ck_2d();

 Wy::cout << "Checked OK\n";
 return(0);
}
catch(const WyRet& e) {
 Wy::cerr << Wy::wrd(e) << "\n";
 return(-1);
}
catch(const std::exception& e) {
 Wy::cerr << "std::exception\n";
 return(-1);
}
catch(...) {
 Wy::cerr << "unknown unwind\n";
 return(-1);
};
