/* Copyright is licensed by GNU LGPL.                         by I.J.Wang 2005
*/
#define WYLIB_SOURCE
#include "wystr.h"
#include "wy__math.h"
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <limits>

static const char WY__EToken('e');

//
// [Internal] Break num into radix notation and store into buf
//
// Note: 1. buf have to have enough space before call
//       2. buf can not be 0, radix should be >=2
//       3. the stored 'digit' is saved in reversed order
//       4. if num is negative, the stored 'digit' is negative
//       5. the stored 'digit' is in range [0-radix)
// 
// [Ret] number of digits stored into buf
//
inline static int wy__brknum(char* buf,int num, int radix) WY__TSPC()
{
 std::div_t dt={num,0};
 //dt.quot=num;
 int n=0;
 do {
   dt=std::div(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline static int wy__brknum(char* buf,long num, int radix) WY__TSPC()
{
 std::ldiv_t dt={num,0};
 //dt.quot=num;
 int n=0;
 do {
   dt=std::ldiv(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline static int wy__brknum(char* buf,long long num, int radix) WY__TSPC()
{
 std::lldiv_t dt={num,0};
 //dt.quot=num;
 int n=0;
 do {
   dt=std::lldiv(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline static int wy__brknum(char* buf,unsigned int num, int radix)
{
 if(num<=(unsigned int)(std::numeric_limits<int>::max())) {
   return( wy__brknum(buf,(int)num,radix) );
 }

 const unsigned int q=num/radix;
 if(q>(unsigned int)(std::numeric_limits<int>::max())) {
   // assertion fail
   WY_THROW(WyRet(Wym_ERANGE));   
 }
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,(int)q,radix)+1 );
};

inline static int wy__brknum(char* buf,unsigned long num, int radix)
{
 if(num<=(unsigned long)(std::numeric_limits<long>::max())) {
   return( wy__brknum(buf,(long)num,radix) );
 }

 const unsigned long q=num/radix;
 if(q>(unsigned long)(std::numeric_limits<long>::max())) {
   // assertion fail
   WY_THROW(WyRet(Wym_ERANGE));   
 }
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,(long)q,radix)+1 );
};

inline static int wy__brknum(char* buf,unsigned long long num, int radix)
{
 if(num<=(unsigned long long)(std::numeric_limits<long long>::max())) {
   return( wy__brknum(buf,(long long)num,radix) );
 }

 const unsigned long long q=num/radix;
 if(q>(unsigned long long)(std::numeric_limits<long long>::max())) {
   // assertion fail
   WY_THROW(WyRet(Wym_ERANGE));   
 }
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,(long long)q,radix)+1 );
};

inline static int wy__brknum(char*,double, int)
{
 // This function exist for compiling reason. should not be actually called
 WY_TERMINATE("");
};

//
// [Internal] Convert the broken down representation in buf/blen
//            to ascii
//
// Note: buf can not be 0, blen must be >0
//
inline static void wy__cvrrev_pos(char* buf, int blen) WY__TSPC()
{
 int i,j;
 for(i=0, j=blen-1; i<=j; ++i,--j) {
   char ch1=buf[i];
   char ch2=buf[j];
   ch1= (ch1>9)? (ch1+'7'):(ch1+'0');
   ch2= (ch2>9)? (ch2+'7'):(ch2+'0');
   buf[i]=ch2;
   buf[j]=ch1;
 }
};

//
// [Internal] Convert the broken down representation in buf/blen
//            to ascii
//
// Note: buf can not be 0, blen must be >0
// Note: buf contaions one more character ('-') after return
//
inline static void wy__cvrrev_neg(char* buf, int blen) WY__TSPC()
{
 {
   const char ch0=-buf[0];
   buf[blen]=(ch0>9)? (ch0+'7'):(ch0+'0');
   buf[0]='-';
 }
 int i,j;
 for(i=1, j=blen-1; i<=j; ++i,--j) {
   char ch1=-buf[i];
   char ch2=-buf[j];
   ch1= (ch1>9)? (ch1+'7'):(ch1+'0');
   ch2= (ch2>9)? (ch2+'7'):(ch2+'0');
   buf[i]=ch2;
   buf[j]=ch1;
 }
};

// [Interna] Convert number(integer type) to string
//
// [Ret] OK
//       Wym_EINVAL      radix invalid
//       Wym_EFBIG       str size exceeds the maximum
//       Wym_ENOMEM      Not enough memory
//
template <typename NumT>
inline WyRet wy__setnum_integer(WyStr &str, NumT num, int radix)
{
 if(std::numeric_limits<NumT>::is_integer==false) {
   WY_TERMINATE("");
 }
 if((radix<2)||(radix>36)) {
   WY_RETURN(Wym_EINVAL);
 }

 char dbuf[std::numeric_limits<NumT>::digits+2]; // digits*('0'|'1') +'-' + 0
 int n=wy__brknum(dbuf,num,radix);
 if(std::numeric_limits<NumT>::is_signed) {
   if(num<0) {
     wy__cvrrev_neg(dbuf,n);
     ++n;
   } else {
     wy__cvrrev_pos(dbuf,n);
   }
 } else {
   wy__cvrrev_pos(dbuf,n);
 }
 WY_RETURN( str.reset( WyCSeg(dbuf,(size_t)n) ) );
};

//  [Syn] Convert number num into radix string
//        representation, reset str with the result. The number of
//        characters for fraction is specified by frdig.
//        [-]ddd.ddd
//
//    Note: There may be truncate or round-up error 
//    
//    Implementation works only for radix equal to 10
//
//  [Ret] OK              Succeed
//        Wym_EINVAL      radix invalid
//        Wym_EFBIG       str size exceeds the maximum
//        Wym_ENOMEM      Not enough memory
//
WyRet wy__setnum_double(WyStr &str,double num,int radix,size_t frdig)
try {
 typedef long long int FI_INT;   // int type capable holding (int)double
 const size_t FI_INT_DIG10( size_t(std::numeric_limits<FI_INT>::digits10) );
 if(radix!=10) {
   WY_RETURN(Wym_EINVAL);
 }

 if(frdig>FI_INT_DIG10) {
   frdig=FI_INT_DIG10;
 }

 if(isnan(num)) {
   str.reset("nan");
   return(OK);
 }
 if(isinf(num)) {
   if(isinf(num)>0) {
     str.reset("inf");
   } else {
     str.reset("-inf");
   }
   return(OK);
 }
 WyRet r;
 WyStr stmp;
 double wnum;   // whole num
 double frac;   // fraction

 if(num<0) {
   stmp.reset(1,'-');   // always OK
   num=-num;
 }
 frac=std::modf(num,&wnum);

 {
   WyStr tmp;
   size_t n_dig( static_cast<size_t>(std::log10(wnum)+1.0) );  // dig10 of wnum
   if(n_dig>=FI_INT_DIG10) {
     wnum/= std::pow((double)10.0,(double)(n_dig-FI_INT_DIG10+1));
   }

   r=wy__setnum_integer(tmp,static_cast<FI_INT>(wnum),10);  
   if(r!=OK) {
     WY_RETURN(r);
   }
   if(n_dig>=FI_INT_DIG10) {
     if((r=tmp.append(n_dig-FI_INT_DIG10+1,'0'))!=OK) {
       WY_RETURN(r);
     }
   }
   stmp+=tmp;
   if(frdig<=0) {
     if(wnum<=0) {
       WY_RETURN( str.reset("0") );
     } else {
       WY_RETURN( str.reset(stmp) );
     }
   }
   stmp+='.';
 }

 std::modf(frac*std::pow(10.0,static_cast<double>(frdig))
           +std::numeric_limits<double>::epsilon()   // roundup?
           ,&wnum);
 {
   WyStr tmp;
   r=wy__setnum_integer(tmp,static_cast<FI_INT>(wnum),10);
   if(r!=OK) {
     WY_RETURN(r);
   }
   if(frdig<tmp.size()) {
     WY_THROW(WyRet(Wym_ERANGE));
   }
   if((r=stmp.append(frdig-tmp.size(),'0'))!=OK) {
     WY_RETURN(r);
   }
   stmp+=tmp;
 }
 str.swap(stmp);
 return(OK);
}
catch(const WyStr::Reply& e) {
 return(e);
};

template <typename NumT>
WyRet wy__setnum(WyStr &str, NumT num, int radix)
try {
 if(std::numeric_limits<NumT>::is_integer) {
   WY_RETURN( wy__setnum_integer(str,num,radix) );
 } else {
   // float NumT assumed
   typedef signed long long int FI_INT;  // int type capable holding (int)double
   const int FrRadix=10;                 // float type always use radix 10
   const size_t DftFracDig=6;            // number of digits for fraction

   if(radix==FrRadix) {
     WY_RETURN(wy__setnum_double(str,num,radix,DftFracDig));
   } else if (radix!=0) {
     WY_RETURN(Wym_EINVAL);
   } else {
     // FALL_THROUGH with radix==0
   }

   double fnorm,p10;
   WyRet r;
   WyStr tmp1;
   fnorm=WyMath::_scif(num,p10);
   if((r=wy__setnum_double(tmp1,fnorm,FrRadix,DftFracDig))!=OK) {
     WY_RETURN(r);
   }

   {
     WyStr tmp2;
     if((r=wy__setnum_integer(tmp2,static_cast<FI_INT>(p10),FrRadix))!=OK) {
       WY_RETURN(r);
     }
     tmp1+=WY__EToken;
     if(p10>=0) {
       tmp1+='+';
     }
     tmp1+=tmp2;
   }
   return( str.reset(tmp1) );
 }
}
catch(const WyStr::Reply& e) {
 WY_RETURN( WyRet(e) );
};

template <typename NumT>
WyRet wy__setnum(WyStr &str, NumT num, int radix, size_t frdig)
try {
 if(std::numeric_limits<NumT>::is_integer) {
   if(frdig!=0) {
     WY_RETURN(Wym_EINVAL);
   }
   WY_RETURN( wy__setnum_integer(str,num,radix) );
 } else {
   // float NumT assumed
   typedef signed long long int FI_INT;  // int type capable holding (int)double
   const int FrRadix=10;                 // float type always use radix 10

   if(radix==FrRadix) {
     WY_RETURN(wy__setnum_double(str,num,radix,frdig));
   } else if (radix!=0) {
     WY_RETURN(Wym_EINVAL);
   } else {
     // FALL_THROUGH with radix==0
   }

   if(isnan(num)) {
     str.reset("nan");
     return(OK);
   }
   if(isinf(num)) {
     if(isinf(num)>0) {
       str.reset("inf");
     } else {
       str.reset("-inf");
     }
     return(OK);
   }
   double fnorm,p10;
   WyRet r;
   WyStr tmp1;
   fnorm=WyMath::_scif(num,p10);
   if((r=wy__setnum_double(tmp1,fnorm,FrRadix,frdig))!=OK) {
     WY_RETURN(r);
   }

   {
     WyStr tmp2;
     if((r=wy__setnum_integer(tmp2,static_cast<FI_INT>(p10),FrRadix))!=OK) {
       WY_RETURN(r);
     }
     tmp1+=WY__EToken;
     if(p10>=0) {
       tmp1+='+';
     }
     tmp1+=tmp2;
   }
   return( str.reset(tmp1) );
 }
}
catch(const WyStr::Reply& e) {
 WY_RETURN( WyRet(e) );
};

template <typename NumT>
WyRet wy__apdnum_integer(WyStr &str, NumT num, int radix)
{
 if((radix<2)||(radix>36)) {
   WY_RETURN(Wym_EINVAL);
 }

 char dbuf[std::numeric_limits<NumT>::digits+2]; // digits*('0'|'1') +'-' + 0
 int n=wy__brknum(dbuf,num,radix);
 if(std::numeric_limits<NumT>::is_signed) {
   if(num<0) {
     wy__cvrrev_neg(dbuf,n);
     ++n;
   } else {
     wy__cvrrev_pos(dbuf,n);
   }
 } else {
   wy__cvrrev_pos(dbuf,n);
 }
 WY_RETURN( str.append( WyCSeg(dbuf,(size_t)n) ) );
};

//  [Syn] Convert number num into radix string
//        representation, append the result to str. The number of
//        characters for fraction is specified by frdig.
//        [-]ddd.ddd
//
//    Note: There may be truncate or round-up error 
//    
//    Implementation works only for radix equal to 10
//
//  [Ret] OK              Succeed
//        Wym_EINVAL      radix invalid
//        Wym_EFBIG       str size exceeds the maximum
//        Wym_ENOMEM      Not enough memory
//
WyRet wy__apdnum_double(WyStr &str,double num,int radix,size_t frdig)
try {
 typedef long long int FI_INT;   // int type capable holding (int)double
 const size_t FI_INT_DIG10( size_t(std::numeric_limits<FI_INT>::digits10) );
 if(radix!=10) {
   WY_RETURN(Wym_EINVAL);
 }

 if(frdig>FI_INT_DIG10) {
   frdig=FI_INT_DIG10;
 }
 
 if(isnan(num)) {
   WY_RETURN( str.append("nan") );
 }
 if(isinf(num)) {
   if(isinf(num)>0) {
     WY_RETURN( str.append("inf") );
   } else {
     WY_RETURN( str.append("-inf") );
   }
 }
 WyRet r;
 WyStr stmp;
 double wnum;   // whole num
 double frac;   // fraction

 if(num<0) {
   stmp.reset(1,'-');   // always OK
   num=-num;
 }
 frac=std::modf(num,&wnum);

 {
   WyStr tmp;
   size_t n_dig( static_cast<size_t>(std::log10(wnum)+1.0) );  // dig10 of wnum
   if(n_dig>=FI_INT_DIG10) {
     wnum/= std::pow((double)10.0,(double)(n_dig-FI_INT_DIG10+1));
   }

   r=wy__setnum_integer(tmp,static_cast<FI_INT>(wnum),10);  
   if(r!=OK) {
     WY_RETURN(r);
   }
   if(n_dig>=FI_INT_DIG10) {
     if((r=tmp.append(n_dig-FI_INT_DIG10+1,'0'))!=OK) {
       WY_RETURN(r);
     }
   }
   stmp+=tmp;
   if(frdig<=0) {
     if(wnum<=0) {
       WY_RETURN( str.append(char('0')) );
     } else {
       WY_RETURN( str.append(stmp) );
     }
   }
   stmp+='.';
 }

 std::modf(frac*std::pow(10.0,static_cast<double>(frdig))
           +std::numeric_limits<double>::epsilon()   // roundup?
           ,&wnum);
 {
   WyStr tmp;
   r=wy__setnum_integer(tmp,static_cast<FI_INT>(wnum),10);
   if(r!=OK) {
     WY_RETURN(r);
   }
   if(frdig<tmp.size()) {
     WY_TERMINATE("");
   }
   if((r=stmp.append(frdig-tmp.size(),'0'))!=OK) {
     WY_RETURN(r);
   }
   stmp+=tmp;
 }
 WY_RETURN( str.append(stmp) );
}
catch(const WyStr::Reply& e) {
 return(e);
};

template <typename NumT>
WyRet wy__apdnum(WyStr &str, NumT num, int radix)
try {
 if(std::numeric_limits<NumT>::is_integer) {
   WY_RETURN( wy__apdnum_integer(str,num,radix) );
 } else {
   // float NumT assumed
   typedef signed long long int FI_INT;  // int type capable holding (int)double
   const int FrRadix=10;                 // float types always use radix 10
   const size_t DftFracDig=6;            // number of digits for fraction

   if(radix==FrRadix) {
     WY_RETURN(wy__apdnum_double(str,num,radix,DftFracDig));
   } else if (radix!=0) {
     WY_RETURN(Wym_EINVAL);
   } else {
     // FALL_THROUGH with radix==0
   }

   if(isnan(num)) {
     str.reset("nan");
     return(OK);
   }
   if(isinf(num)) {
     if(isinf(num)>0) {
       str.reset("inf");
     } else {
       str.reset("-inf");
     }
     return(OK);
   }
   double fnorm,p10;
   WyRet r;
   WyStr tmp1;
   fnorm=WyMath::_scif(num,p10);
   if((r=wy__setnum_double(tmp1,fnorm,FrRadix,DftFracDig))!=OK) {
     WY_RETURN(r);
   }

   {
     WyStr tmp2;
     if((r=wy__setnum_integer(tmp2,static_cast<FI_INT>(p10),FrRadix))!=OK) {
       WY_RETURN(r);
     }
     tmp1+=WY__EToken;
     if(p10>=0) {
       tmp1+='+';
     }
     tmp1+=tmp2;
   }
   return( str.append(tmp1) );
 }
}
catch(const WyStr::Reply& e) {
 return(e);
};

template <typename NumT>
WyRet wy__apdnum(WyStr &str, NumT num, int radix, size_t frdig)
try {
 if(std::numeric_limits<NumT>::is_integer) {
   if(frdig!=0) {
     WY_RETURN(Wym_EINVAL);
   }
   WY_RETURN( wy__apdnum_integer(str,num,radix) );
 } else {
   // float NumT assumed
   typedef signed long long int FI_INT;  // int type capable holding (int)double
   const int FrRadix=10;                 // float types always use radix 10

   if(radix==FrRadix) {
     WY_RETURN(wy__apdnum_double(str,num,radix,frdig));
   } else if (radix!=0) {
     WY_RETURN(Wym_EINVAL);
   } else {
     // FALL_THROUGH with radix==0
   }

   double fnorm,p10;
   WyRet r;
   WyStr tmp1;
   fnorm=WyMath::_scif(num,p10);
   if((r=wy__setnum_double(tmp1,fnorm,FrRadix,frdig))!=OK) {
     WY_RETURN(r);
   }

   {
     WyStr tmp2;
     if((r=wy__setnum_integer(tmp2,static_cast<FI_INT>(p10),FrRadix))!=OK) {
       WY_RETURN(r);
     }
     tmp1+=WY__EToken;
     if(p10>=0) {
       tmp1+='+';
     }
     tmp1+=tmp2;
   }
   return( str.append(tmp1) );
 }
}
catch(const WyStr::Reply& e) {
 return(e);
};

namespace Wy {

WyRet _mkstr(WyStr& str, char num)
{ return( wy__setnum(str,(int)num,10) ); };

WyRet _mkstr(WyStr& str, signed char num)
{ return( wy__setnum(str,(signed int)num,10) ); };

WyRet _mkstr(WyStr& str, signed short num)
{ return( wy__setnum(str,(signed int)num,10) ); };

WyRet _mkstr(WyStr& str, signed int num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, signed long num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, signed long long num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, unsigned char num)
{ return( wy__setnum(str,(unsigned int)num,10) ); };

WyRet _mkstr(WyStr& str, unsigned short num)
{ return( wy__setnum(str,(unsigned int)num,10) ); };

WyRet _mkstr(WyStr& str, unsigned int num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, unsigned long num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, unsigned long long num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, float num)
{ return( wy__setnum(str,num,10) ); };

WyRet _mkstr(WyStr& str, double num)
{ return( wy__setnum(str,num,10) ); };

//
WyRet _mkstr(WyStr& str, char num,int radix)
{ return( wy__setnum(str,(int)num,radix) ); };

WyRet _mkstr(WyStr& str, signed char num,int radix)
{ return( wy__setnum(str,(signed int)num,radix) ); };

WyRet _mkstr(WyStr& str, signed short num,int radix)
{ return( wy__setnum(str,(signed int)num,radix) ); };

WyRet _mkstr(WyStr& str, signed int num,int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, signed long num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, signed long long num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr& str, unsigned char num,int radix)
{ return( wy__setnum(str,(unsigned int)num,radix) ); };

WyRet _mkstr(WyStr& str, unsigned short num,int radix)
{ return( wy__setnum(str,(unsigned int)num,radix) ); };

WyRet _mkstr(WyStr &str, unsigned int num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, unsigned long num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, unsigned long long num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, float num, int radix)
{ return( wy__setnum(str,num,radix) ); };

WyRet _mkstr(WyStr &str, double num, int radix)
{ return( wy__setnum(str,num,radix) ); };

//--------
WyRet _mkstr(WyStr &str, char num, int radix, size_t frdig)
{ return( wy__setnum(str,(int)num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, signed char num, int radix, size_t frdig)
{ return( wy__setnum(str,(signed int)num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, signed short num, int radix, size_t frdig)
{ return( wy__setnum(str,(signed int)num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, signed int num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, signed long num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, signed long long num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, unsigned char num, int radix, size_t frdig)
{ return( wy__setnum(str,(unsigned int)num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, unsigned short num, int radix, size_t frdig)
{ return( wy__setnum(str,(unsigned int)num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, unsigned int num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, unsigned long num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, unsigned long long num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, float num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

WyRet _mkstr(WyStr &str, double num, int radix, size_t frdig)
{ return( wy__setnum(str,num,radix,frdig) ); };

//---
 WyRet _catstr(WyStr& str, char num)
{ return wy__apdnum(str,(int)num,10); };

WyRet _catstr(WyStr& str, signed char num)
{ return wy__apdnum(str,(signed int)num,10); };

WyRet _catstr(WyStr& str, signed short num)
{ return wy__apdnum(str,(signed int)num,10); };

WyRet _catstr(WyStr& str, signed int num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, signed long num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, signed long long num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, unsigned char num)
{ return wy__apdnum(str,(unsigned int)num,10); };

WyRet _catstr(WyStr& str, unsigned short num)
{ return wy__apdnum(str,(unsigned int)num,10); };

WyRet _catstr(WyStr& str, unsigned int num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, unsigned long num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, unsigned long long num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, float num)
{ return wy__apdnum(str,num,10); };

WyRet _catstr(WyStr& str, double num)
{ return wy__apdnum(str,num,10); };

//--
WyRet _catstr(WyStr &str, char num, int radix)
{ return wy__apdnum(str,(int)num,radix); };

WyRet _catstr(WyStr &str, signed char num, int radix)
{ return wy__apdnum(str,(signed int)num,radix); };

WyRet _catstr(WyStr &str, signed short num, int radix)
{ return wy__apdnum(str,(signed int)num,radix); };

WyRet _catstr(WyStr &str, signed int num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, signed long num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, signed long long num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, unsigned char num, int radix)
{ return wy__apdnum(str,(unsigned int)num,radix); };

WyRet _catstr(WyStr &str, unsigned short num, int radix)
{ return wy__apdnum(str,(unsigned int)num,radix); };

WyRet _catstr(WyStr &str, unsigned int num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, unsigned long num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, unsigned long long num, int radix)
{ return wy__apdnum(str,num,radix); };

WyRet _catstr(WyStr &str, float num, int radix)
{ return( wy__apdnum(str,num,radix) ); };

WyRet _catstr(WyStr &str, double num, int radix)
{ return( wy__apdnum(str,num,radix) ); };

//---
WyRet _catstr(WyStr &str, char num, int radix, size_t frdig)
{ return wy__apdnum(str,(int)num,radix,frdig); };

WyRet _catstr(WyStr &str, signed char num, int radix, size_t frdig)
{ return wy__apdnum(str,(signed int)num,radix,frdig); };

WyRet _catstr(WyStr &str, signed short num, int radix, size_t frdig)
{ return wy__apdnum(str,(signed int)num,radix,frdig); };

WyRet _catstr(WyStr &str, signed int num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, signed long num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, signed long long num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, unsigned char num, int radix, size_t frdig)
{ return wy__apdnum(str,(unsigned int)num,radix,frdig); };

WyRet _catstr(WyStr &str, unsigned short num, int radix, size_t frdig)
{ return wy__apdnum(str,(unsigned int)num,radix,frdig); };

WyRet _catstr(WyStr &str, unsigned int num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, unsigned long num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, unsigned long long num, int radix, size_t frdig)
{ return wy__apdnum(str,num,radix,frdig); };

WyRet _catstr(WyStr &str, float num, int radix, size_t frdig)
{ return( wy__apdnum(str,num,radix,frdig) ); };

WyRet _catstr(WyStr &str, double num, int radix, size_t frdig)
{ return( wy__apdnum(str,num,radix,frdig) ); };

//--------------------
//WyStr wrd(char num)
//{
// return WyStr(size_t(1),num);
//};

WyStr wrd(char num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,(int)num,radix) );
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed char num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,(signed int)num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed short num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,(signed int)num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed int num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed long num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed long long num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned char num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,(unsigned int)num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned short num,int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,(unsigned int)num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned int num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned long num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned long long num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(float num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(double num, int radix)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

//--------------------------
WyStr wrd(char num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,(int)num,radix,frdig) );
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed char num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,(signed int)num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed short num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,(signed int)num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed int num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed long num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(signed long long num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned char num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,(unsigned int)num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned short num,int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,(unsigned int)num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned int num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned long num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(unsigned long long num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(float num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

WyStr wrd(double num, int radix, size_t frdig)
{
 WyStr str;
 WyRet r( wy__setnum(str,num,radix,frdig) ); 
 if(r!=OK) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

}; // End Wy

